from urllib.parse import urlparse

import pytest
import yaml
from dagster._core.test_utils import ensure_dagster_tests_import, environ, instance_for_test
from dagster_mysql.run_storage import MySQLRunStorage

ensure_dagster_tests_import()
from dagster_tests.storage_tests.utils.run_storage import TestRunStorage

TestRunStorage.__test__ = False


class TestMySQLRunStorage(TestRunStorage):
    __test__ = True
    # TestMySQLRunStorage::test_backfill_tags_filtering_multiple_results

    def supports_backfill_tags_filtering_queries(self) -> bool:
        return True

    def supports_backfill_job_name_filtering_queries(self) -> bool:
        return True

    def supports_backfill_id_filtering_queries(self) -> bool:
        return True

    def supports_backfills_count(self) -> bool:
        return True

    @pytest.fixture(name="instance", scope="function")
    def instance(self, conn_string):
        MySQLRunStorage.create_clean_storage(conn_string)

        with instance_for_test(
            overrides={"storage": {"mysql": {"mysql_url": conn_string}}}
        ) as instance:
            yield instance

    @pytest.fixture(scope="function", name="storage")
    def run_storage(self, instance):
        run_storage = instance.run_storage
        assert isinstance(run_storage, MySQLRunStorage)
        return run_storage

    def test_load_from_config(self, conn_string):
        parse_result = urlparse(conn_string)
        hostname = parse_result.hostname  # can be custom set in the BK env
        port = (
            parse_result.port
        )  # can be different, based on the backcompat mysql version or latest mysql version

        url_cfg = f"""
          run_storage:
            module: dagster_mysql.run_storage
            class: MySQLRunStorage
            config:
              mysql_url: mysql+mysqlconnector://test:test@{hostname}:{port}/test
        """

        explicit_cfg = f"""
          run_storage:
            module: dagster_mysql.run_storage
            class: MySQLRunStorage
            config:
              mysql_db:
                username: test
                password: test
                hostname: {hostname}
                db_name: test
                port: {port}
        """

        with environ({"TEST_MYSQL_PASSWORD": "test"}):
            env_cfg = f"""
            run_storage:
              module: dagster_mysql.run_storage
              class: MySQLRunStorage
              config:
                mysql_db:
                  username: test
                  password:
                    env: TEST_MYSQL_PASSWORD
                  hostname: {hostname}
                  db_name: test
                  port: {port}
            """

            with instance_for_test(overrides=yaml.safe_load(url_cfg)) as from_url_instance:
                with instance_for_test(
                    overrides=yaml.safe_load(explicit_cfg)
                ) as from_explicit_instance:
                    assert (
                        from_url_instance._run_storage.mysql_url  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
                        == from_explicit_instance._run_storage.mysql_url  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
                    )
                with instance_for_test(overrides=yaml.safe_load(env_cfg)) as from_env_instance:
                    assert (
                        from_url_instance._run_storage.mysql_url  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
                        == from_env_instance._run_storage.mysql_url  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
                    )



import time

import pytest
import sqlalchemy as db
from dagster_mysql.run_storage import MySQLRunStorage


def retry_connect(conn_string: str, num_retries: int = 5, pool_recycle=-1):
    storage = MySQLRunStorage.create_clean_storage(conn_string)
    storage.optimize_for_webserver(-1, pool_recycle=pool_recycle, max_overflow=20)

    with storage.connect() as conn:
        conn.execute(db.text("SET SESSION wait_timeout = 2;"))

    for _ in range(num_retries):
        time.sleep(3)
        with storage.connect() as conn:
            conn.execute(db.text("SELECT 1;"))

    return storage.get_runs()


def test_pool_recycle_greater_than_wait_timeout(conn_string):
    with pytest.raises(db.exc.OperationalError):  # pyright: ignore[reportAttributeAccessIssue]
        retry_connect(conn_string)


def test_pool_recycle_less_than_wait_timeout(conn_string):
    runs_lst = retry_connect(conn_string, pool_recycle=1)
    assert len(runs_lst) == 0



# ruff: noqa: SLF001
import datetime
import os
import subprocess
import tempfile
from urllib.parse import urlparse

import pytest
import sqlalchemy as db
from dagster import AssetKey, AssetMaterialization, AssetObservation, Output, job, op
from dagster._core.definitions.data_version import DATA_VERSION_TAG
from dagster._core.errors import DagsterInvalidInvocationError
from dagster._core.events import DagsterEvent, DagsterEventType
from dagster._core.execution.backfill import BulkActionsFilter, BulkActionStatus, PartitionBackfill
from dagster._core.instance import DagsterInstance
from dagster._core.remote_representation.external_data import partition_set_snap_name_for_job_name
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus, RunsFilter
from dagster._core.storage.event_log.migration import ASSET_KEY_INDEX_COLS
from dagster._core.storage.migration.bigint_migration import run_bigint_migration
from dagster._core.storage.runs.migration import BACKFILL_JOB_NAME_AND_TAGS, RUN_BACKFILL_ID
from dagster._core.storage.sqlalchemy_compat import db_select
from dagster._core.storage.tags import BACKFILL_ID_TAG
from dagster._core.utils import make_new_run_id
from dagster._daemon.types import DaemonHeartbeat
from dagster._time import get_current_timestamp
from dagster._utils import file_relative_path


def get_columns(instance, table_name: str):
    with instance.run_storage.connect() as conn:
        return set(c["name"] for c in db.inspect(conn).get_columns(table_name))


def get_indexes(instance, table_name: str):
    with instance.run_storage.connect() as conn:
        return set(i["name"] for i in db.inspect(conn).get_indexes(table_name))


def get_tables(instance):
    with instance.run_storage.connect() as conn:
        return db.inspect(conn).get_table_names()


def _reconstruct_from_file(conn_string, path, _username="root", _password="test"):
    parse_result = urlparse(conn_string)
    hostname = parse_result.hostname
    port = parse_result.port
    engine = db.create_engine(conn_string)
    with engine.connect() as conn:
        with conn.begin():
            conn.execute(db.text("drop schema test;"))
            conn.execute(db.text("create schema test;"))
    env = os.environ.copy()
    env["MYSQL_PWD"] = "test"
    subprocess.check_call(
        f"mysql -uroot -h{hostname} -P{port} -ptest test < {path}", shell=True, env=env
    )
    return hostname, port


def test_0_13_17_mysql_convert_float_cols(conn_string):
    hostname, port = _reconstruct_from_file(
        conn_string,
        file_relative_path(__file__, "snapshot_0_13_18_start_end_timestamp.sql"),
    )

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        instance = DagsterInstance.from_config(tempdir)
        record = instance.get_run_records(limit=1)[0]
        assert int(record.start_time) == 1643760000  # pyright: ignore[reportArgumentType]
        assert int(record.end_time) == 1643760000  # pyright: ignore[reportArgumentType]

        instance.upgrade()

        record = instance.get_run_records(limit=1)[0]
        assert record.start_time is None
        assert record.end_time is None

        instance.reindex()

        record = instance.get_run_records(limit=1)[0]
        assert int(record.start_time) == 1643788829  # pyright: ignore[reportArgumentType]
        assert int(record.end_time) == 1643788834  # pyright: ignore[reportArgumentType]


def test_instigators_table_backcompat(conn_string):
    hostname, port = _reconstruct_from_file(
        conn_string,
        file_relative_path(__file__, "snapshot_0_14_6_instigators_table.sql"),
    )

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        instance = DagsterInstance.from_config(tempdir)

        assert not instance.schedule_storage.has_instigators_table()  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]

        instance.upgrade()

        assert instance.schedule_storage.has_instigators_table()  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]


def test_asset_observation_backcompat(conn_string):
    hostname, port = _reconstruct_from_file(
        conn_string,
        file_relative_path(__file__, "snapshot_0_11_16_pre_add_asset_key_index_cols.sql"),
    )

    @op
    def asset_op(_):
        yield AssetObservation(asset_key=AssetKey(["a"]))
        yield Output(1)

    @job
    def asset_job():
        asset_op()

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        with DagsterInstance.from_config(tempdir) as instance:
            storage = instance._event_storage

            assert not instance.event_log_storage.has_secondary_index(ASSET_KEY_INDEX_COLS)  # pyright: ignore[reportAttributeAccessIssue]

            asset_job.execute_in_process(instance=instance)
            assert storage.has_asset_key(AssetKey(["a"]))


def test_jobs_selector_id_migration(conn_string):
    import sqlalchemy as db
    from dagster._core.storage.schedules.migration import SCHEDULE_JOBS_SELECTOR_ID
    from dagster._core.storage.schedules.schema import InstigatorsTable, JobTable, JobTickTable

    hostname, port = _reconstruct_from_file(
        conn_string,
        file_relative_path(__file__, "snapshot_0_14_6_post_schema_pre_data_migration.sql"),
    )

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        with DagsterInstance.from_config(tempdir) as instance:
            # runs the required data migrations
            instance.upgrade()

            assert instance.schedule_storage.has_built_index(SCHEDULE_JOBS_SELECTOR_ID)  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]
            legacy_count = len(instance.all_instigator_state())
            migrated_instigator_count = instance.schedule_storage.execute(  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]
                db_select([db.func.count()]).select_from(InstigatorsTable)
            )[0][0]
            assert migrated_instigator_count == legacy_count

            migrated_job_count = instance.schedule_storage.execute(  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]
                db_select([db.func.count()])
                .select_from(JobTable)
                .where(JobTable.c.selector_id.isnot(None))
            )[0][0]
            assert migrated_job_count == legacy_count

            legacy_tick_count = instance.schedule_storage.execute(  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]
                db_select([db.func.count()]).select_from(JobTickTable)
            )[0][0]
            assert legacy_tick_count > 0

            # tick migrations are optional
            migrated_tick_count = instance.schedule_storage.execute(  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]
                db_select([db.func.count()])
                .select_from(JobTickTable)
                .where(JobTickTable.c.selector_id.isnot(None))
            )[0][0]
            assert migrated_tick_count == 0

            # run the optional migrations
            instance.reindex()

            migrated_tick_count = instance.schedule_storage.execute(  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]
                db_select([db.func.count()])
                .select_from(JobTickTable)
                .where(JobTickTable.c.selector_id.isnot(None))
            )[0][0]
            assert migrated_tick_count == legacy_tick_count


def test_add_bulk_actions_columns(conn_string):
    new_columns = {"selector_id", "action_type"}
    new_indexes = {"idx_bulk_actions_action_type", "idx_bulk_actions_selector_id"}

    hostname, port = _reconstruct_from_file(
        conn_string,
        # use an old snapshot, it has the bulk actions table but not the new columns
        file_relative_path(__file__, "snapshot_0_14_6_post_schema_pre_data_migration.sql"),
    )

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        with DagsterInstance.from_config(tempdir) as instance:
            assert get_columns(instance, "bulk_actions") & new_columns == set()
            assert get_indexes(instance, "bulk_actions") & new_indexes == set()

            instance.upgrade()
            assert new_columns <= get_columns(instance, "bulk_actions")
            assert new_indexes <= get_indexes(instance, "bulk_actions")


def test_add_kvs_table(conn_string):
    hostname, port = _reconstruct_from_file(
        conn_string,
        # use an old snapshot
        file_relative_path(__file__, "snapshot_0_14_6_post_schema_pre_data_migration.sql"),
    )

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        with DagsterInstance.from_config(tempdir) as instance:
            assert "kvs" not in get_tables(instance)

            instance.upgrade()
            assert "kvs" in get_tables(instance)
            assert "idx_kvs_keys_unique" in get_indexes(instance, "kvs")


def test_add_asset_event_tags_table(conn_string):
    @op
    def yields_materialization_w_tags(_):
        yield AssetMaterialization(asset_key=AssetKey(["a"]), tags={DATA_VERSION_TAG: "bar"})
        yield Output(1)

    @job
    def asset_job():
        yields_materialization_w_tags()

    hostname, port = _reconstruct_from_file(
        conn_string,
        # use an old snapshot
        file_relative_path(__file__, "snapshot_1_0_12_pre_add_asset_event_tags_table.sql"),
    )

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        with DagsterInstance.from_config(tempdir) as instance:
            assert "asset_event_tags" not in get_tables(instance)
            asset_job.execute_in_process(instance=instance)
            with pytest.raises(
                DagsterInvalidInvocationError, match="In order to search for asset event tags"
            ):
                instance._event_storage.get_event_tags_for_asset(asset_key=AssetKey(["a"]))

            instance.upgrade()
            assert "asset_event_tags" in get_tables(instance)
            asset_job.execute_in_process(instance=instance)
            assert instance._event_storage.get_event_tags_for_asset(asset_key=AssetKey(["a"])) == [
                {DATA_VERSION_TAG: "bar"}
            ]

            indexes = get_indexes(instance, "asset_event_tags")
            assert "idx_asset_event_tags" in indexes
            assert "idx_asset_event_tags_event_id" in indexes


def test_add_cached_status_data_column(conn_string):
    new_columns = {"cached_status_data"}

    hostname, port = _reconstruct_from_file(
        conn_string,
        # use an old snapshot, it has the bulk actions table but not the new columns
        file_relative_path(__file__, "snapshot_1_0_17_add_cached_status_data_column.sql"),
    )

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        with DagsterInstance.from_config(tempdir) as instance:
            assert get_columns(instance, "asset_keys") & new_columns == set()

            instance.upgrade()
            assert new_columns <= get_columns(instance, "asset_keys")


def test_add_dynamic_partitions_table(conn_string):
    hostname, port = _reconstruct_from_file(
        conn_string,
        file_relative_path(__file__, "snapshot_1_0_17_add_cached_status_data_column.sql"),
    )

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        with DagsterInstance.from_config(tempdir) as instance:
            assert "dynamic_partitions" not in get_tables(instance)

            instance.wipe()

            with pytest.raises(DagsterInvalidInvocationError, match="does not exist"):
                instance.get_dynamic_partitions("foo")

            instance.upgrade()
            assert "dynamic_partitions" in get_tables(instance)
            assert instance.get_dynamic_partitions("foo") == []


def _get_table_row_count(run_storage, table, with_non_null_id=False):
    import sqlalchemy as db

    query = db_select([db.func.count()]).select_from(table)
    if with_non_null_id:
        query = query.where(table.c.id.isnot(None))
    with run_storage.connect() as conn:
        row_count = conn.execute(query).fetchone()[0]
    return row_count


def test_add_primary_keys(conn_string):
    from dagster._core.storage.runs.schema import (
        DaemonHeartbeatsTable,
        InstanceInfo,
        KeyValueStoreTable,
    )

    hostname, port = _reconstruct_from_file(
        conn_string,
        file_relative_path(__file__, "snapshot_1_1_22_pre_primary_key.sql"),
    )

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        with DagsterInstance.from_config(tempdir) as instance:
            assert "id" not in get_columns(instance, "kvs")
            # trigger insert, and update
            instance.run_storage.set_cursor_values({"a": "A"})
            instance.run_storage.set_cursor_values({"a": "A"})

            kvs_row_count = _get_table_row_count(instance.run_storage, KeyValueStoreTable)
            assert kvs_row_count > 0

            assert "id" not in get_columns(instance, "instance_info")
            instance_info_row_count = _get_table_row_count(instance.run_storage, InstanceInfo)
            assert instance_info_row_count > 0

            assert "id" not in get_columns(instance, "daemon_heartbeats")
            heartbeat = DaemonHeartbeat(
                timestamp=datetime.datetime.now().timestamp(), daemon_type="test", daemon_id="test"
            )
            instance.run_storage.add_daemon_heartbeat(heartbeat)
            instance.run_storage.add_daemon_heartbeat(heartbeat)
            daemon_heartbeats_row_count = _get_table_row_count(
                instance.run_storage, DaemonHeartbeatsTable
            )
            assert daemon_heartbeats_row_count > 0

            instance.upgrade()

            assert "id" in get_columns(instance, "kvs")
            with instance.run_storage.connect():  # pyright: ignore[reportAttributeAccessIssue]
                kvs_id_count = _get_table_row_count(
                    instance.run_storage, KeyValueStoreTable, with_non_null_id=True
                )
            assert kvs_id_count == kvs_row_count

            assert "id" in get_columns(instance, "instance_info")
            with instance.run_storage.connect():  # pyright: ignore[reportAttributeAccessIssue]
                instance_info_id_count = _get_table_row_count(
                    instance.run_storage, InstanceInfo, with_non_null_id=True
                )
            assert instance_info_id_count == instance_info_row_count

            assert "id" in get_columns(instance, "daemon_heartbeats")
            with instance.run_storage.connect():  # pyright: ignore[reportAttributeAccessIssue]
                daemon_heartbeats_id_count = _get_table_row_count(
                    instance.run_storage, DaemonHeartbeatsTable, with_non_null_id=True
                )
            assert daemon_heartbeats_id_count == daemon_heartbeats_row_count


def test_bigint_migration(conn_string):
    hostname, port = _reconstruct_from_file(
        conn_string,
        file_relative_path(__file__, "snapshot_1_1_22_pre_primary_key.sql"),
    )

    def _get_integer_id_tables(conn):
        inspector = db.inspect(conn)
        integer_tables = set()
        for table in inspector.get_table_names():
            type_by_col_name = {c["name"]: c["type"] for c in db.inspect(conn).get_columns(table)}
            id_type = type_by_col_name.get("id")
            if id_type and str(id_type) == "INTEGER":
                integer_tables.add(table)
        return integer_tables

    def _assert_autoincrement_id(conn):
        inspector = db.inspect(conn)
        for table in inspector.get_table_names():
            id_cols = [col for col in inspector.get_columns(table) if col["name"] == "id"]
            if id_cols:
                id_col = id_cols[0]
                assert id_col["autoincrement"]

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        with DagsterInstance.from_config(tempdir) as instance:
            with instance.run_storage.connect() as conn:  # pyright: ignore[reportAttributeAccessIssue]
                assert len(_get_integer_id_tables(conn)) > 0
                _assert_autoincrement_id(conn)
            with instance.event_log_storage.index_connection() as conn:  # pyright: ignore[reportAttributeAccessIssue]
                assert len(_get_integer_id_tables(conn)) > 0
                _assert_autoincrement_id(conn)
            with instance.schedule_storage.connect() as conn:  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]
                assert len(_get_integer_id_tables(conn)) > 0
                _assert_autoincrement_id(conn)

            run_bigint_migration(instance)

            with instance.run_storage.connect() as conn:  # pyright: ignore[reportAttributeAccessIssue]
                assert len(_get_integer_id_tables(conn)) == 0
                _assert_autoincrement_id(conn)
            with instance.event_log_storage.index_connection() as conn:  # pyright: ignore[reportAttributeAccessIssue]
                assert len(_get_integer_id_tables(conn)) == 0
                _assert_autoincrement_id(conn)
            with instance.schedule_storage.connect() as conn:  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]
                assert len(_get_integer_id_tables(conn)) == 0
                _assert_autoincrement_id(conn)


def test_add_backfill_id_column(conn_string):
    from dagster._core.storage.runs.schema import RunsTable

    new_columns = {"backfill_id"}

    hostname, port = _reconstruct_from_file(
        conn_string,
        # use an old snapshot, it has the bulk actions table but not the new columns
        file_relative_path(
            __file__, "snapshot_1_8_12_pre_add_backfill_id_column_to_runs_table.sql"
        ),
    )

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        with DagsterInstance.from_config(tempdir) as instance:
            assert get_columns(instance, "runs") & new_columns == set()

            # these runs won't have an entry for backfill_id until after the data migration
            run_not_in_backfill_pre_migration = instance.run_storage.add_run(
                DagsterRun(
                    job_name="first_job_no_backfill",
                    run_id=make_new_run_id(),
                    tags=None,
                    status=DagsterRunStatus.NOT_STARTED,
                )
            )
            run_in_backfill_pre_migration = instance.run_storage.add_run(
                DagsterRun(
                    job_name="first_job_in_backfill",
                    run_id=make_new_run_id(),
                    tags={BACKFILL_ID_TAG: "backfillid"},
                    status=DagsterRunStatus.NOT_STARTED,
                )
            )

            # exclude_subruns filter works before migration
            assert len(instance.get_runs(filters=RunsFilter(exclude_subruns=True))) == 2

            instance.upgrade()
            assert instance.run_storage.has_built_index(RUN_BACKFILL_ID)  # pyright: ignore[reportAttributeAccessIssue]
            assert new_columns <= get_columns(instance, "runs")

            run_not_in_backfill_post_migration = instance.run_storage.add_run(
                DagsterRun(
                    job_name="second_job_no_backfill",
                    run_id=make_new_run_id(),
                    tags=None,
                    status=DagsterRunStatus.NOT_STARTED,
                )
            )
            run_in_backfill_post_migration = instance.run_storage.add_run(
                DagsterRun(
                    job_name="second_job_in_backfill",
                    run_id=make_new_run_id(),
                    tags={BACKFILL_ID_TAG: "backfillid"},
                    status=DagsterRunStatus.NOT_STARTED,
                )
            )

            backfill_ids = {
                row["run_id"]: row["backfill_id"]
                for row in instance._run_storage.fetchall(  # pyright: ignore[reportAttributeAccessIssue]
                    db_select([RunsTable.c.run_id, RunsTable.c.backfill_id]).select_from(RunsTable)
                )
            }
            assert backfill_ids[run_not_in_backfill_pre_migration.run_id] is None
            assert backfill_ids[run_in_backfill_pre_migration.run_id] == "backfillid"
            assert backfill_ids[run_not_in_backfill_post_migration.run_id] is None
            assert backfill_ids[run_in_backfill_post_migration.run_id] == "backfillid"
            # exclude_subruns filter works after migration, but should use new column
            assert len(instance.get_runs(filters=RunsFilter(exclude_subruns=True))) == 3


def test_add_runs_by_backfill_id_idx(conn_string):
    hostname, port = _reconstruct_from_file(
        conn_string,
        file_relative_path(
            __file__, "snapshot_1_8_12_pre_add_backfill_id_column_to_runs_table.sql"
        ),
    )

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        with DagsterInstance.from_config(tempdir) as instance:
            assert get_indexes(instance, "runs") & {"idx_runs_by_backfill_id"} == set()
            instance.upgrade()
            assert {"idx_runs_by_backfill_id"} <= get_indexes(instance, "runs")


def test_add_backfill_tags(conn_string):
    from dagster._core.storage.runs.schema import BackfillTagsTable

    hostname, port = _reconstruct_from_file(
        conn_string,
        # use an old snapshot, it has the bulk actions table but not the new columns
        file_relative_path(
            __file__, "snapshot_1_8_12_pre_add_backfill_id_column_to_runs_table.sql"
        ),
    )

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        with DagsterInstance.from_config(tempdir) as instance:
            assert "backfill_tags" not in get_tables(instance)
            before_migration = PartitionBackfill(
                "before_tag_migration",
                serialized_asset_backfill_data="foo",
                status=BulkActionStatus.REQUESTED,
                from_failure=False,
                tags={"before": "migration"},
                backfill_timestamp=get_current_timestamp(),
            )
            instance.add_backfill(before_migration)
            # filtering pre-migration relies on filtering runs, so add a run with the expected tags
            pre_migration_run = instance.run_storage.add_run(
                DagsterRun(
                    job_name="foo",
                    run_id=make_new_run_id(),
                    tags={"before": "migration", BACKFILL_ID_TAG: before_migration.backfill_id},
                    status=DagsterRunStatus.NOT_STARTED,
                )
            )

            # filtering by tags works before migration
            assert (
                instance.get_backfills(filters=BulkActionsFilter(tags={"before": "migration"}))[
                    0
                ].backfill_id
                == before_migration.backfill_id
            )

            instance.upgrade()
            assert "backfill_tags" in get_tables(instance)

            after_migration = PartitionBackfill(
                "after_tag_migration",
                serialized_asset_backfill_data="foo",
                status=BulkActionStatus.REQUESTED,
                from_failure=False,
                tags={"after": "migration"},
                backfill_timestamp=get_current_timestamp(),
            )
            instance.add_backfill(after_migration)

            with instance.run_storage.connect() as conn:  # pyright: ignore[reportAttributeAccessIssue]
                rows = conn.execute(
                    db_select(
                        [
                            BackfillTagsTable.c.backfill_id,
                            BackfillTagsTable.c.key,
                            BackfillTagsTable.c.value,
                        ]
                    )
                ).fetchall()
                assert len(rows) == 2
                ids_to_tags = {row[0]: {row[1]: row[2]} for row in rows}
                assert ids_to_tags.get(before_migration.backfill_id) == before_migration.tags
                assert ids_to_tags[after_migration.backfill_id] == after_migration.tags

                # filtering by tags works after migration
                assert instance.run_storage.has_built_index(BACKFILL_JOB_NAME_AND_TAGS)  # pyright: ignore[reportAttributeAccessIssue]
                # delete the run that was added pre-migration to prove that tags filtering is happening on the
                # backfill_tags table
                instance.delete_run(pre_migration_run.run_id)
                assert (
                    instance.get_backfills(filters=BulkActionsFilter(tags={"before": "migration"}))[
                        0
                    ].backfill_id
                    == before_migration.backfill_id
                )
                assert (
                    instance.get_backfills(filters=BulkActionsFilter(tags={"after": "migration"}))[
                        0
                    ].backfill_id
                    == after_migration.backfill_id
                )


def test_add_bulk_actions_job_name_column(conn_string):
    from dagster._core.remote_origin import (
        GrpcServerCodeLocationOrigin,
        RemotePartitionSetOrigin,
        RemoteRepositoryOrigin,
    )
    from dagster._core.storage.runs.schema import BulkActionsTable

    hostname, port = _reconstruct_from_file(
        conn_string,
        # use an old snapshot, it has the bulk actions table but not the new columns
        file_relative_path(
            __file__, "snapshot_1_8_12_pre_add_backfill_id_column_to_runs_table.sql"
        ),
    )
    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        with DagsterInstance.from_config(tempdir) as instance:
            assert "job_name" not in get_columns(instance, "bulk_actions")
            partition_set_origin = RemotePartitionSetOrigin(
                repository_origin=RemoteRepositoryOrigin(
                    code_location_origin=GrpcServerCodeLocationOrigin(
                        host="localhost", port=1234, location_name="test_location"
                    ),
                    repository_name="the_repo",
                ),
                partition_set_name=partition_set_snap_name_for_job_name("before_migration"),
            )
            before_migration = PartitionBackfill(
                "before_migration",
                partition_set_origin=partition_set_origin,
                status=BulkActionStatus.REQUESTED,
                from_failure=False,
                tags={},
                backfill_timestamp=get_current_timestamp(),
            )
            instance.add_backfill(before_migration)
            # filtering pre-migration relies on filtering runs, so add a run with the expected job_name
            pre_migration_run = instance.run_storage.add_run(
                DagsterRun(
                    job_name=before_migration.job_name,  # pyright: ignore[reportArgumentType]
                    run_id=make_new_run_id(),
                    tags={BACKFILL_ID_TAG: before_migration.backfill_id},
                    status=DagsterRunStatus.NOT_STARTED,
                )
            )

            # filtering by job_name works before migration
            assert (
                instance.get_backfills(
                    filters=BulkActionsFilter(job_name=before_migration.job_name)
                )[0].backfill_id
                == before_migration.backfill_id
            )

            instance.upgrade()

            assert "job_name" in get_columns(instance, "bulk_actions")

            partition_set_origin = RemotePartitionSetOrigin(
                repository_origin=RemoteRepositoryOrigin(
                    code_location_origin=GrpcServerCodeLocationOrigin(
                        host="localhost", port=1234, location_name="test_location"
                    ),
                    repository_name="the_repo",
                ),
                partition_set_name=partition_set_snap_name_for_job_name("after_migration"),
            )
            after_migration = PartitionBackfill(
                "after_migration",
                partition_set_origin=partition_set_origin,
                status=BulkActionStatus.REQUESTED,
                from_failure=False,
                tags={},
                backfill_timestamp=get_current_timestamp(),
            )
            instance.add_backfill(after_migration)

            with instance.run_storage.connect() as conn:  # pyright: ignore[reportAttributeAccessIssue]
                rows = conn.execute(
                    db_select([BulkActionsTable.c.key, BulkActionsTable.c.job_name])
                ).fetchall()
                assert len(rows) == 2
                ids_to_job_name = {row[0]: row[1] for row in rows}
                assert ids_to_job_name[before_migration.backfill_id] == before_migration.job_name
                assert ids_to_job_name[after_migration.backfill_id] == after_migration.job_name

                # filtering by job_name works after migration
                assert instance.run_storage.has_built_index(BACKFILL_JOB_NAME_AND_TAGS)  # pyright: ignore[reportAttributeAccessIssue]
                # delete the run that was added pre-migration to prove that tags filtering is happening on the
                # backfill_tags table
                instance.delete_run(pre_migration_run.run_id)
                assert (
                    instance.get_backfills(
                        filters=BulkActionsFilter(job_name=before_migration.job_name)
                    )[0].backfill_id
                    == before_migration.backfill_id
                )
                assert (
                    instance.get_backfills(
                        filters=BulkActionsFilter(job_name=after_migration.job_name)
                    )[0].backfill_id
                    == after_migration.backfill_id
                )


def test_add_run_tags_run_id_idx(conn_string):
    hostname, port = _reconstruct_from_file(
        conn_string,
        # use an old snapshot, it has the bulk actions table but not the new columns
        file_relative_path(
            __file__, "snapshot_1_8_12_pre_add_backfill_id_column_to_runs_table.sql"
        ),
    )

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        with DagsterInstance.from_config(tempdir) as instance:
            # Before migration
            assert "run_tags" in get_tables(instance)
            assert "idx_run_tags" in get_indexes(instance, "run_tags")
            assert "idx_run_tags_run_id" not in get_indexes(instance, "run_tags")

            # After upgrade
            instance.upgrade()

            assert "run_tags" in get_tables(instance)
            assert "idx_run_tags" not in get_indexes(instance, "run_tags")
            assert "idx_run_tags_run_id" in get_indexes(instance, "run_tags")


def test_add_backfill_end_timestamp(conn_string):
    hostname, port = _reconstruct_from_file(
        conn_string,
        # use an old snapshot, it has the bulk actions table but not the new columns
        file_relative_path(__file__, "snapshot_1_9_3_add_run_tags_run_id_idx.sql"),
    )

    with tempfile.TemporaryDirectory() as tempdir:
        with open(file_relative_path(__file__, "dagster.yaml"), encoding="utf8") as template_fd:
            with open(os.path.join(tempdir, "dagster.yaml"), "w", encoding="utf8") as target_fd:
                template = template_fd.read().format(hostname=hostname, port=port)
                target_fd.write(template)

        with DagsterInstance.from_config(tempdir) as instance:
            completed_backfill = PartitionBackfill(
                "before_end_timestamp_migration",
                serialized_asset_backfill_data="foo",
                status=BulkActionStatus.COMPLETED_SUCCESS,
                from_failure=False,
                tags={},
                backfill_timestamp=get_current_timestamp(),
            )
            instance.add_backfill(completed_backfill)
            in_progress_backfill = PartitionBackfill(
                "in_progress_backfill",
                serialized_asset_backfill_data="foo",
                status=BulkActionStatus.REQUESTED,
                from_failure=False,
                tags={},
                backfill_timestamp=get_current_timestamp(),
            )
            instance.add_backfill(in_progress_backfill)
            no_runs_backfill = PartitionBackfill(
                "no_runs_backfill",
                serialized_asset_backfill_data="foo",
                status=BulkActionStatus.CANCELED,
                from_failure=False,
                tags={},
                backfill_timestamp=get_current_timestamp(),
            )
            instance.add_backfill(no_runs_backfill)
            # before migration, backfill end timestamps will be None
            completed_backfill = instance.get_backfill(completed_backfill.backfill_id)
            assert completed_backfill
            assert completed_backfill.backfill_end_timestamp is None
            in_progress_backfill = instance.get_backfill(in_progress_backfill.backfill_id)
            assert in_progress_backfill
            assert in_progress_backfill.backfill_end_timestamp is None
            no_runs_backfill = instance.get_backfill(no_runs_backfill.backfill_id)
            assert no_runs_backfill
            assert no_runs_backfill.backfill_end_timestamp is None

            for _ in range(3):
                instance.run_storage.add_run(
                    DagsterRun(
                        job_name="foo",
                        run_id=make_new_run_id(),
                        tags={BACKFILL_ID_TAG: completed_backfill.backfill_id},
                        status=DagsterRunStatus.NOT_STARTED,
                    )
                )
                instance.run_storage.add_run(
                    DagsterRun(
                        job_name="foo",
                        run_id=make_new_run_id(),
                        tags={BACKFILL_ID_TAG: in_progress_backfill.backfill_id},
                        status=DagsterRunStatus.NOT_STARTED,
                    )
                )
            completed_backfill_run_end_times = []
            for run in instance.get_runs(
                filters=RunsFilter.for_backfill(completed_backfill.backfill_id)
            ):
                dagster_event = DagsterEvent(
                    event_type_value=DagsterEventType.RUN_SUCCESS.value,
                    job_name=run.job_name,
                    message="yay run success",
                    step_key="bar",
                )
                instance.report_dagster_event(dagster_event, run_id=run.run_id)
                updated_run = instance.get_run_record_by_id(run.run_id)
                completed_backfill_run_end_times.append(
                    updated_run.end_time if updated_run else None
                )

            for run in instance.get_runs(
                filters=RunsFilter.for_backfill(in_progress_backfill.backfill_id)
            ):
                dagster_event = DagsterEvent(
                    event_type_value=DagsterEventType.RUN_SUCCESS.value,
                    job_name=run.job_name,
                    message="yay run success",
                    step_key="bar",
                )
                instance.report_dagster_event(dagster_event, run_id=run.run_id)

            instance.upgrade()

            completed_backfill = instance.get_backfill(completed_backfill.backfill_id)
            assert completed_backfill
            assert (
                completed_backfill.backfill_end_timestamp is not None
                and completed_backfill.backfill_end_timestamp
                == max(completed_backfill_run_end_times)
            )

            in_progress_backfill = instance.get_backfill(in_progress_backfill.backfill_id)
            assert in_progress_backfill
            assert in_progress_backfill.backfill_end_timestamp is None

            no_runs_backfill = instance.get_backfill(no_runs_backfill.backfill_id)
            assert no_runs_backfill
            assert no_runs_backfill.backfill_end_timestamp == no_runs_backfill.backfill_timestamp



from urllib.parse import urlparse

import pytest
from dagster._utils import file_relative_path
from dagster._utils.test.mysql_instance import TestMySQLInstance


@pytest.fixture(scope="session")
def hostname(conn_string):
    parse_result = urlparse(conn_string)
    return parse_result.hostname


@pytest.fixture(scope="session")
def conn_string():
    with TestMySQLInstance.docker_service_up_or_skip(
        file_relative_path(__file__, "docker-compose.yml"), "test-mysql-db"
    ) as conn_str:
        yield conn_str



from dagster_shared.libraries import DagsterLibraryRegistry

from dagster_mysql.event_log import MySQLEventLogStorage
from dagster_mysql.run_storage import MySQLRunStorage
from dagster_mysql.schedule_storage import MySQLScheduleStorage
from dagster_mysql.storage import DagsterMySQLStorage
from dagster_mysql.version import __version__

DagsterLibraryRegistry.register("dagster-mysql", __version__)
__all__ = ["DagsterMySQLStorage", "MySQLEventLogStorage", "MySQLRunStorage", "MySQLScheduleStorage"]

from dagster_mysql.resources import MySQLResource as MySQLResource



from urllib.parse import urlparse

from dagster import asset, materialize_to_memory
from dagster_mysql import MySQLResource


def test_resource(hostname, conn_string):
    @asset
    def mysql_create_table(mysql: MySQLResource):
        with mysql.get_connection() as conn:
            with conn.cursor() as cur:
                cur.execute(
                    "CREATE TABLE IF NOT EXISTS films ("
                    " title VARCHAR(40) NOT NULL,"
                    " year INT NOT NULL, "
                    " PRIMARY KEY (title));"
                )
                cur.execute(
                    "INSERT INTO films (title, year)"
                    " VALUES ('the matrix', 1999)"
                    " ON DUPLICATE KEY UPDATE title=title;"
                )
                cur.execute(
                    "INSERT INTO films (title, year)"
                    " VALUES ('fight club', 1999)"
                    " ON DUPLICATE KEY UPDATE title=title;"
                )

            conn.commit()

    @asset
    def mysql_query_table(mysql: MySQLResource):
        with mysql.get_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("SELECT * FROM films;")
                assert len(cur.fetchall()) == 2

    conn_info = urlparse(conn_string)

    result = materialize_to_memory(
        [mysql_create_table, mysql_query_table],
        resources={
            "mysql": MySQLResource(
                user=conn_info.username,
                password=conn_info.password,
                host=conn_info.hostname,
                port=conn_info.port,
                database="test",
            )
        },
    )

    assert result.success



import gc
import time
from contextlib import contextmanager
from urllib.parse import urlparse

import objgraph
import pytest
import yaml
from dagster._core.storage.event_log.base import EventLogCursor
from dagster._core.test_utils import ensure_dagster_tests_import, instance_for_test
from dagster._core.utils import make_new_run_id
from dagster_mysql.event_log import MySQLEventLogStorage

ensure_dagster_tests_import()
from dagster_tests.storage_tests.utils.event_log_storage import (
    TestEventLogStorage,
    create_test_event_log_record,
)


@contextmanager
def _clean_storage(conn_string):
    storage = MySQLEventLogStorage.create_clean_storage(conn_string)
    assert storage
    try:
        yield storage
    finally:
        storage.dispose()


class TestMySQLEventLogStorage(TestEventLogStorage):
    __test__ = True

    @pytest.fixture(name="instance", scope="function")
    def instance(self, conn_string):
        MySQLEventLogStorage.create_clean_storage(conn_string)

        with instance_for_test(
            overrides={"storage": {"mysql": {"mysql_url": conn_string}}}
        ) as instance:
            yield instance

    def can_wipe_asset_partitions(self) -> bool:
        return False

    @pytest.fixture(scope="function", name="storage")
    def event_log_storage(self, instance):
        event_log_storage = instance.event_log_storage
        assert isinstance(event_log_storage, MySQLEventLogStorage)
        yield event_log_storage

    def test_event_log_storage_two_watchers(self, conn_string):
        with _clean_storage(conn_string) as storage:
            run_id = make_new_run_id()
            watched_1 = []
            watched_2 = []

            def watch_one(event, _cursor):
                watched_1.append(event)

            def watch_two(event, _cursor):
                watched_2.append(event)

            assert len(storage.get_logs_for_run(run_id)) == 0

            storage.store_event(create_test_event_log_record(str(1), run_id=run_id))
            assert len(storage.get_logs_for_run(run_id)) == 1
            assert len(watched_1) == 0

            storage.watch(run_id, str(EventLogCursor.from_storage_id(1)), watch_one)

            storage.store_event(create_test_event_log_record(str(2), run_id=run_id))
            storage.store_event(create_test_event_log_record(str(3), run_id=run_id))

            storage.watch(run_id, str(EventLogCursor.from_storage_id(3)), watch_two)
            storage.store_event(create_test_event_log_record(str(4), run_id=run_id))

            attempts = 10
            while (len(watched_1) < 3 or len(watched_2) < 1) and attempts > 0:
                time.sleep(0.1)
                attempts -= 1

            assert len(storage.get_logs_for_run(run_id)) == 4
            assert len(watched_1) == 3
            assert len(watched_2) == 1

            storage.end_watch(run_id, watch_one)
            time.sleep(0.3)  # this value scientifically selected from a range of attractive values
            storage.store_event(create_test_event_log_record(str(5), run_id=run_id))

            attempts = 10
            while len(watched_2) < 2 and attempts > 0:
                time.sleep(0.1)
                attempts -= 1
            storage.end_watch(run_id, watch_two)

            assert len(storage.get_logs_for_run(run_id)) == 5
            assert len(watched_1) == 3
            assert len(watched_2) == 2

            storage.delete_events(run_id)

            assert len(storage.get_logs_for_run(run_id)) == 0
            assert len(watched_1) == 3
            assert len(watched_2) == 2

            assert [int(evt.message) for evt in watched_1] == [2, 3, 4]
            assert [int(evt.message) for evt in watched_2] == [4, 5]
            assert len(objgraph.by_type("SqlPollingEventWatcher")) == 1

        # ensure we clean up poller on exit
        gc.collect()
        assert len(objgraph.by_type("SqlPollingEventWatcher")) == 0

    def test_load_from_config(self, conn_string):
        parse_result = urlparse(conn_string)
        hostname = parse_result.hostname  # can be custom set in the BK env
        port = (
            parse_result.port
        )  # can be different, based on the backcompat mysql version or latest mysql version

        url_cfg = f"""
        event_log_storage:
            module: dagster_mysql.event_log
            class: MySQLEventLogStorage
            config:
                mysql_url: mysql+mysqlconnector://test:test@{hostname}:{port}/test
        """

        explicit_cfg = f"""
        event_log_storage:
            module: dagster_mysql.event_log
            class: MySQLEventLogStorage
            config:
                mysql_db:
                    username: test
                    password: test
                    hostname: {hostname}
                    port: {port}
                    db_name: test
        """

        with instance_for_test(overrides=yaml.safe_load(url_cfg)) as from_url_instance:
            from_url = from_url_instance._event_storage  # noqa: SLF001

            with instance_for_test(overrides=yaml.safe_load(explicit_cfg)) as explicit_instance:
                from_explicit = explicit_instance._event_storage  # noqa: SLF001

                assert from_url.mysql_url == from_explicit.mysql_url  # pyright: ignore[reportAttributeAccessIssue]



from dagster import _check as check
from dagster._config.config_schema import UserConfigSchema
from dagster._core.storage.base_storage import DagsterStorage
from dagster._core.storage.config import MySqlStorageConfig, mysql_config
from dagster._core.storage.event_log import EventLogStorage
from dagster._core.storage.runs import RunStorage
from dagster._core.storage.schedules import ScheduleStorage
from dagster._serdes import ConfigurableClass, ConfigurableClassData

from dagster_mysql.event_log import MySQLEventLogStorage
from dagster_mysql.run_storage import MySQLRunStorage
from dagster_mysql.schedule_storage import MySQLScheduleStorage
from dagster_mysql.utils import mysql_url_from_config


class DagsterMySQLStorage(DagsterStorage, ConfigurableClass):
    """MySQL-backed dagster storage.

    Users should not directly instantiate this class; it is instantiated by internal machinery when
    ``dagster-webserver`` and ``dagster-graphql`` load, based on the values in the ``dagster.yaml`` file in
    ``$DAGSTER_HOME``. Configuration of this class should be done by setting values in that file.

    To use MySQL for storage, you can add a block such as the following to your
    ``dagster.yaml``:

    .. literalinclude:: ../../../../../../examples/docs_snippets/docs_snippets/deploying/dagster-mysql.yaml
       :caption: dagster.yaml
       :language: YAML

    Note that the fields in this config are :py:class:`~dagster.StringSource` and
    :py:class:`~dagster.IntSource` and can be configured from environment variables.
    """

    def __init__(self, mysql_url, inst_data: ConfigurableClassData | None = None):
        self.mysql_url = mysql_url
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        self._run_storage = MySQLRunStorage(mysql_url)
        self._event_log_storage = MySQLEventLogStorage(mysql_url)
        self._schedule_storage = MySQLScheduleStorage(mysql_url)
        super().__init__()

    @property
    def inst_data(self) -> ConfigurableClassData | None:
        return self._inst_data

    @classmethod
    def config_type(cls) -> UserConfigSchema:
        return mysql_config()

    @classmethod
    def from_config_value(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls, inst_data: ConfigurableClassData | None, config_value: MySqlStorageConfig
    ) -> "DagsterMySQLStorage":
        return DagsterMySQLStorage(
            inst_data=inst_data,
            mysql_url=mysql_url_from_config(config_value),
        )

    @property
    def event_log_storage(self) -> EventLogStorage:
        return self._event_log_storage

    @property
    def run_storage(self) -> RunStorage:
        return self._run_storage

    @property
    def schedule_storage(self) -> ScheduleStorage:
        return self._schedule_storage

    @property
    def event_storage_data(self) -> ConfigurableClassData | None:
        return (
            ConfigurableClassData(
                "dagster_mysql",
                "MySQLEventLogStorage",
                self.inst_data.config_yaml,
            )
            if self.inst_data
            else None
        )

    @property
    def run_storage_data(self) -> ConfigurableClassData | None:
        return (
            ConfigurableClassData(
                "dagster_mysql",
                "MySQLRunStorage",
                self.inst_data.config_yaml,
            )
            if self.inst_data
            else None
        )

    @property
    def schedule_storage_data(self) -> ConfigurableClassData | None:
        return (
            ConfigurableClassData(
                "dagster_mysql",
                "MySQLScheduleStorage",
                self.inst_data.config_yaml,
            )
            if self.inst_data
            else None
        )



import pytest
from dagster._core.test_utils import ensure_dagster_tests_import
from dagster_mysql.run_storage import MySQLRunStorage

ensure_dagster_tests_import()
from dagster_tests.storage_tests.utils.daemon_cursor_storage import TestDaemonCursorStorage


class TestMySQLDaemonCursorStorage(TestDaemonCursorStorage):
    __test__ = True

    @pytest.fixture(scope="function", name="storage")
    def cursor_storage(self, conn_string):
        storage = MySQLRunStorage.create_clean_storage(conn_string)
        assert storage
        return storage



from tempfile import TemporaryDirectory
from urllib.parse import urlparse

import pytest
import sqlalchemy as db
import yaml
from dagster._core.instance import DagsterInstance
from dagster._core.instance.ref import InstanceRef
from dagster._core.storage.sql import create_engine, get_alembic_config, stamp_alembic_rev
from dagster._core.test_utils import instance_for_test
from dagster._utils import file_relative_path
from dagster_mysql import MySQLEventLogStorage, MySQLRunStorage, MySQLScheduleStorage
from dagster_mysql.utils import mysql_isolation_level
from sqlalchemy.pool import NullPool


def full_mysql_config(hostname, port):
    return f"""
      run_storage:
        module: dagster_mysql.run_storage
        class: MySQLRunStorage
        config:
          mysql_db:
            username: test
            password: test
            hostname: {hostname}
            port: {port}
            db_name: test

      event_log_storage:
        module: dagster_mysql.event_log
        class: MySQLEventLogStorage
        config:
            mysql_db:
              username: test
              password: test
              hostname: {hostname}
              port: {port}
              db_name: test

      schedule_storage:
        module: dagster_mysql.schedule_storage
        class: MySQLScheduleStorage
        config:
            mysql_db:
              username: test
              password: test
              hostname: {hostname}
              port: {port}
              db_name: test
    """


def unified_mysql_config(hostname, port):
    return f"""
      storage:
        mysql:
          mysql_db:
            username: test
            password: test
            hostname: {hostname}
            db_name: test
            port: {port}
    """


def test_connection_leak(conn_string):
    parse_result = urlparse(conn_string)
    hostname = parse_result.hostname
    port = parse_result.port

    num_instances = 20

    tempdir = TemporaryDirectory()
    copies = []
    for _ in range(num_instances):
        copies.append(
            DagsterInstance.from_ref(
                InstanceRef.from_dir(
                    tempdir.name, overrides=yaml.safe_load(full_mysql_config(hostname, port))
                )
            )
        )

    engine = create_engine(conn_string, isolation_level=mysql_isolation_level(), poolclass=NullPool)
    with engine.connect() as conn:
        with conn.begin():
            row = conn.execute(
                db.text("SELECT count(*) FROM information_schema.processlist")
            ).fetchone()

    # This includes a number of internal connections, so just ensure it did not scale
    # with number of instances
    assert row[0] < num_instances  # pyright: ignore[reportOperatorIssue,reportOptionalSubscript]

    for copy in copies:
        copy.dispose()

    tempdir.cleanup()


def test_load_instance(conn_string):
    parse_result = urlparse(conn_string)
    hostname = parse_result.hostname
    port = parse_result.port

    # Wipe the DB to ensure it is fresh
    MySQLEventLogStorage.wipe_storage(conn_string)
    MySQLRunStorage.wipe_storage(conn_string)
    MySQLScheduleStorage.wipe_storage(conn_string)
    engine = create_engine(conn_string, poolclass=NullPool)
    alembic_config = get_alembic_config(
        file_relative_path(__file__, "../dagster_mysql/__init__.py")
    )
    with engine.connect() as conn:
        stamp_alembic_rev(alembic_config, conn, rev=None)  # pyright: ignore[reportArgumentType]

    # Now load from scratch, verify it loads without errors
    with instance_for_test(overrides=yaml.safe_load(full_mysql_config(hostname, port))):
        pass

    # Now load from scratch, using unified storage config
    with instance_for_test(overrides=yaml.safe_load(unified_mysql_config(hostname, port))):
        pass


@pytest.mark.skip("https://github.com/dagster-io/dagster/issues/3719")
def test_statement_timeouts(conn_string):
    parse_result = urlparse(conn_string)
    hostname = parse_result.hostname
    port = parse_result.port

    with instance_for_test(overrides=yaml.safe_load(full_mysql_config(hostname, port))) as instance:
        instance.optimize_for_webserver(
            statement_timeout=500, pool_recycle=-1, max_overflow=20
        )  # 500ms

        # ensure migration error is not raised by being up to date
        instance.upgrade()

        with pytest.raises(db.exc.OperationalError, match="QueryCanceled"):  # pyright: ignore[reportAttributeAccessIssue]
            with instance._run_storage.connect() as conn:  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
                conn.execute(db.text("select sleep(1)")).fetchone()

        with pytest.raises(db.exc.OperationalError, match="QueryCanceled"):  # pyright: ignore[reportAttributeAccessIssue]
            with instance._event_storage.connect() as conn:  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
                conn.execute(db.text("select sleep(1)")).fetchone()

        with pytest.raises(db.exc.OperationalError, match="QueryCanceled"):  # pyright: ignore[reportAttributeAccessIssue]
            with instance._schedule_storage.connect() as conn:  # noqa: SLF001  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]
                conn.execute(db.text("select sleep(1)")).fetchone()
